# -*- Mode: Python; tab-width: 4 -*-
#	Author: Sam Rushing <rushing@nightmare.com>

# Build an auto-attach frozen-module dll (.pfl)

# When the dll is loaded (usually through a call to LoadLibrary),
# DllMain will glue the array of frozen modules into the python
# exported variable PyImport_FrozenModules.

# A pfl file is a normal DLL file, with extra information appended to
# it (this seems harmless), which is used by the 'freeze' program to
# identify which python modules are frozen in it.  [the last 4 bytes
# of the file identify the size of this information]

import os
import string
import sys
import findmodules
import makefreeze
import marshal
import struct

def main (python_src, libname, module_paths):
	if not os.path.isdir ('./build_pfl'):
		os.mkdir ('./build_pfl', 755)
	fo = open ('./build_pfl/frozen.c', 'w')
	d = {}
	for path in module_paths:
		dir, file = os.path.split (path)
		mod, ext = os.path.splitext (file)
		if ext != '.py':
			print 'Not a python file? ("%s")' % path
			continue
		d[mod] = path
	makefreeze.makefreeze (fo, d)
	fo.close()
	open ('./build_pfl/frozenmain.cpp', 'w').write (frozenmain_cpp)
	open ('./build_pfl/makefile', 'w').write (makefile_template % (python_src, libname))
	os.chdir ('./build_pfl')
	os.system ('nmake')
	ds = marshal.dumps([d])
	fdo = open ('%s.pfl' % libname, 'ab')
	fdo.write (ds)
	fdo.write (struct.pack ('l', len(ds)))
	fdo.close()

def get_pfl_directory (filename):
	fd = open (filename, 'rb')
	fd.seek (-4, 2)
	dlen, = struct.unpack ('l', fd.read (4))
	if dlen < 10 or dlen > 10000:
		raise ValueError, "%s is probably not a pfl file (dlen == %d)" % (filename, dlen)
	fd.seek (-(4 + dlen), 2)
	# using a list allows us to expand the format later
	info_list = marshal.load (fd)
	# the directory is the first element
	d = info_list[0]
	fd.close()
	return d

makefile_template = \
"""
pythondir=%s
module=%s

PCH = 0
BUILD_DL=1
!include <$(pythondir)\make_nt.in>

lcustom=
srcdir=$(pythondir)

all: $(module).pfl

$(module).pfl : frozenmain.obj
	link -dll -out:$*.pfl frozenmain.obj $(lcustom) $(python_defs_lib)
"""

frozenmain_cpp = \
"""
#include "Python.h"
#include "import.h"
#include "frozen.c"

struct _frozen * new_list;

extern __declspec (dllimport) struct _frozen * PyImport_FrozenModules;

static int
glue_frozen (void)
{
  struct _frozen * p, *q;
  int nmods = 0;

  // how many modules in the existing list?
  for (p = PyImport_FrozenModules; p->name; p++) {
	nmods++;
  }
  // how many in the new list?
  for (p = _frozen_modules; p->name; p++) {
	nmods++;
  }
  new_list = (struct _frozen *) malloc (sizeof (struct _frozen) * (nmods + 1));
  if (!new_list) {
	return -1;
  }
  q = new_list;
  // copy the old list in
  for (p = PyImport_FrozenModules; p->name; p++, q++) {
	q->name = p->name;
	q->code = p->code;
	q->size = p->size;
  }
  // copy the new list in
  for (p = _frozen_modules; p->name; p++, q++) {
	q->name = p->name;
	q->code = p->code;
	q->size = p->size;
  }
  // tack on the new sentinel
  q->name = 0;
  q->code = 0;
  q->size = 0;
  // plug in the new list
  PyImport_FrozenModules = new_list;
  return 0;
}

#include <windows.h>

BOOL WINAPI
DllMain (HANDLE hInst, 
	 ULONG ul_reason_for_call,
	 LPVOID lpReserved)
{
  switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:
	  if (glue_frozen () < 0) {
		return FALSE;
	  }
      break;
    case DLL_PROCESS_DETACH:
	  free (new_list);
      break;
    }
  return TRUE;
}

"""

def get_dir_py_files (dirname):
	files = filter (
		lambda x: (os.path.splitext (x)[1] == '.py'),
		os.listdir (dirname)
		)
	return map (
		lambda x,d=dirname: os.path.join (d, x),
		files
		)

def build_directory_dll (python_src, libname, paths):
	files = []
	for path in paths:
		if os.path.isdir (path):
			files = files + get_dir_py_files (path)
		else:
			files.append (path)
	main (python_src, libname, files)
	
if __name__ == '__main__':
	if len(sys.argv) < 3:
		print 'Usage: pfl <python_src_directory> <libname> dirname | filename.py [...]'
	else:
		build_directory_dll (sys.argv[1], sys.argv[2], sys.argv[3:])
