# -*- Mode: Python; tab-width: 4 -*-

import ihooks
import marshal
import bsddb
import strop

databases = {}

import os

def compile_files_into_db (db, files):
	for file in files:
		dir, fname = os.path.split (file)
		name, ext = os.path.splitext (fname)
		print 'compiling %s...' % name
		if ext == '.py':
			try:
				code = compile (
					open(file, 'r').read(),
					file,
					'exec'
					)
			except:
				print 'failed compiling "%s"' % file
				continue
			db[name] = marshal.dumps (code)
		else:
			print 'What is "%s"??' % file
	db.sync()

def get_dir_py_files (dirname):
	files = filter (
		lambda x: (os.path.splitext (x)[1] == '.py'),
		os.listdir (dirname)
		)
	return map (
		lambda x,d=dirname: os.path.join (d, x),
		files
		)

def compile_paths (libname, paths):
	files = []
	for path in paths:
		if os.path.isdir (path):
			files = files + get_dir_py_files (path)
		else:
			files.append (path)
	db = bsddb.btopen (libname+'.db', 'c')
	compile_files_into_db (db, files)
	db.close()

class box_loader (ihooks.ModuleLoader):
	def find_module_in_dir (self, name, dir):
		if dir is not None and (strop.find (dir, '.db') != -1):
			# dir is really a pyc database file
			if databases.has_key (dir):
				db = databases[dir]
			else:
				try:
					db = bsddb.btopen (dir, 'rw')
				except bsddb.error:
					db = bsddb.btopen (dir, 'c')
				databases[dir] = db
			if db.has_key (name):
				return db, dir, ('', '', 'BOX')
		return ihooks.ModuleLoader.find_module_in_dir (self, name, dir)	
			
	def load_module (self, name, stuff):
		db, dbname, info = stuff
		suff, mode, type = info
		if type == 'BOX':
			code = marshal.loads (db[name])
			m = self.hooks.add_module (name)
			exec code in m.__dict__
			return m
		else:
			return ihooks.ModuleLoader.load_module (self, name, stuff)
			

class box_importer (ihooks.ModuleImporter):
	def __init__ (self, loader = None, verbose = 0):
		ihooks.ModuleImporter.__init__ (
			self,
			loader or box_loader(None, verbose),
			verbose
			)

def install (v=0):
	ihooks.install (box_importer (None, v))

def uninstall ():
	ihooks.uninstall()

install()

if __name__ == '__main__':
	import sys
	if len(sys.argv) < 3:
		print 'Usage: pycbox libname (file.py | directory) [...]'
		sys.exit(0)
	else:
		compile_paths (sys.argv[1], sys.argv[2:])
